
/********************************************************************/
/*  ANIMAL - A pseudo artifical intelligence program developed to   */
/*   provide examples of the RAM-BASE virtual dynamic memory 		*/
/*   allocation routines.  Because Animal uses ALL of the Ram-Base  */
/*   functions, you can not link animal.obj with the shareware      */
/*   version of RAM-BASE.  An exe file is included in the shareware */
/*   package, or you can register for the development version of    */
/*   RAM-BASE.														*/
/********************************************************************/


/********************************************************************/
/* This program builds a knowlege base interactively.  The knowlege */
/*  is stored in a tree composed of 'ANIMAL_NODE' structures.  The  */
/*  leaf nodes of this tree are all possible answers, and the 		*/
/*  interior nodes of the tree are all questions to differentiate   */
/*  between the possible answers.  The program works by traversing  */
/*  the tree of questions until it reaches a leaf node hypothesis.  */
/*  If the hypothesis is correct, the program wins.  If the guess   */
/*  is wrong the program adds a new question node, places the old   */
/*  hypothesis as one leaf of the question and adds the new animal  */
/*  as the other leaf (a new hypothesis for the next turn).			*/
/********************************************************************/


/********************************************************************/
/*								INCLUDE FILES						*/
/********************************************************************/
#define LINT_ARGS  1

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <process.h>
#include <assert.h>

#include "RAM-BASE.H"    /* RAM_BASE definitions */


/********************************************************************/
/*						GLOBAL TYPES								*/
/********************************************************************/
typedef struct animal_node
{
    char        q_a;        /* is it a question or an answer */
    char        text[80];   /* what is the question/answer   */

/*  the next two fields apply only to questions  */
    RB_PTR  yes;			/* where do I go if they answered yes */
    RB_PTR  no;				/* where do I go if they answered no  */
}
    ANIMAL_NODE;

#define QUESTION 'Q'   /* each node is either a question or */
#define ANSWER   'A'   /* an answer (a guess) */


/********************************************************************/
/*						GLOBAL MACROS								*/
/********************************************************************/
/* since rbdemand returns a (char *), I use this macro to type cast */
int ier;
#define RB_ANIMAL(X) ((ANIMAL_NODE *) (rbdemand (schema, X, &ier)))
#define IERTEST if (ier != 0) {printf ("ier %d\n", ier); assert (ier==0);}
/********************************************************************/
/*						GLOBAL VARIABLES							*/
/********************************************************************/
char copyright[] = "Copyright Ron Albury 1987";
RB_SCHEMA schema;      /* The infamous & pervasive schema variable. */



/********************************************************************/
/*						SOLVE FUNCTION								*/
/********************************************************************/
/* A recursive function which traverses the knowlege base until it	*/
/*  finds an answer or gives up.  If the knowlege base gets too big */
/*  you would want to make this non-recursive.						*/
/********************************************************************/
void solve (RB_PTR *);
void solve (node)
RB_PTR *node;
{
    char    resp[80];      /* General string variable for responses */


    printf ("\n\n\n\nPlease answer with Yes or No \n\n");

   /** If this is a leaf (answer) node, then make the guess. **/
    if (RB_ANIMAL (node)->q_a == ANSWER)
    {
        printf ("Is it a(n) %s ? ", RB_ANIMAL (node)->text);
        IERTEST
        gets (resp);

       /** Either I win **/
        if (toupper (resp[0]) == 'Y')
        {
            printf ("I win  - thanks for the game \n\n\n");
            return;
        }
        else /** Or I learn **/
        {  /*
            * This part of the function adds two nodes to the tree.  It
            * keeps the old answer in one of the nodes and puts the
            * new answer in the other.  It then converts the old guess
            * node into a question node and attaches the two new nodes
            * to their respective branches.
           */

            RB_PTR keep_node, new_node;

           /** The two additional nodes are both answers. **/
            keep_node = rballoc (schema, sizeof (ANIMAL_NODE), &ier);
            IERTEST
            new_node = rballoc (schema, sizeof (ANIMAL_NODE), &ier);
            IERTEST
            RB_ANIMAL (&keep_node)->q_a = ANSWER;
            IERTEST
            RB_ANIMAL (&new_node)->q_a = ANSWER;
            IERTEST
           /** Set the text field of the two additional nodes. **/
            strcpy (RB_ANIMAL (&keep_node)->text, RB_ANIMAL (node)->text);
            printf ("I give up -- what is it ? ");
            gets (RB_ANIMAL (&new_node)->text);
            IERTEST
          /** Convert the old node to a question. **/
            RB_ANIMAL (node)->q_a = QUESTION;
            IERTEST
            printf ("\nPlease give me a Y/N question to ask next time \n");
            gets (RB_ANIMAL (node)->text);
            IERTEST
          /** Hook the new nodes on their respective branches. **/
            printf ("For %s, would you answer yes or no \n",
             RB_ANIMAL (&new_node)->text);
            IERTEST
            gets (resp);
            if (toupper (resp[0]) == 'Y')
            {
               RB_ANIMAL (node)->yes = new_node;
               IERTEST
               RB_ANIMAL (node)->no = keep_node;
               IERTEST
            }
            else
            {
               RB_ANIMAL (node)->yes = keep_node;
               IERTEST
               RB_ANIMAL (node)->no = new_node;
               IERTEST
            }
            return;
        }
    }
    else /** Else this is a question node - so ask the question **/
    {
        printf ("%s ? ", RB_ANIMAL (node)->text);
        gets (resp);

       /** If the answer is yes, follow the yes branch of the tree. **/
        if (toupper (resp[0]) == 'Y')
            solve (&(RB_ANIMAL (node)->yes));
        else  /** If not, follow the no branch of the tree. **/
            solve (&(RB_ANIMAL (node)->no));
    }
}



/********************************************************************/
/*						MAIN FUNCTION								*/
/********************************************************************/
/* A very simple main routine.  Initialize the RAM_BASE, then call  */
/*  the solution routine with the root node of the tree each time   */
/*  the person wants to play.  Also provides examples of committing */
/*  and rolling back the RAM_BASE.									*/
/********************************************************************/
main ()
{
    int         ier;
    char        resp[80];      /* General string variable for responses. */
    RB_PTR      animal_root;   /* All pointers are RB_POINTER. */


/** STARTUP CODE **/
    printf ("start from scratch (Yes/No) ? ");
    gets (resp);
    if (toupper (resp[0]) == 'Y')
    {
       /** Call rbinit and tell it to create a new swap file. **/
        ier = rbcreate (&schema, 1024, 18, "ANIMAL.DAT", &animal_root,
            sizeof (ANIMAL_NODE));
        if (ier != 0)
        {
            printf ("rbcreate error %d\n", ier);
            exit(0);
        }
       /** The first guess will always be a toad. **/
        RB_ANIMAL (&animal_root)->q_a = ANSWER;
        IERTEST
        strcpy (RB_ANIMAL (&animal_root)->text, "Toad ");
    }
    else
    {
       /** Run off of the old file **/
        ier = rbopen (&schema, 18, "ANIMAL.DAT", &animal_root);
        if (ier != 0)
        {
            printf ("rbopen error %d\n", ier);
            exit(0);
        }
    }

/** PLAYING THE GAME **/
    do  /** Each time through this loop is one game. **/
    {
       printf ("\n\nThink of an animal and I will try to guess\n\n");
       solve (&(animal_root));
        do
        {
            printf ("\n\n Commit/Rollback/Again/Stop ? ");
            gets (resp);
            if (toupper(resp[0]) == 'C')
                rbcommit(schema);
            else if (toupper(resp[0]) == 'R')
                rbrollback(schema);
        }
        while ((toupper(resp[0]) != 'S') &&
               (toupper(resp[0]) != 'A'));

    }
    while (toupper(resp[0]) != 'S');

}